#include <Arduino.h>
#include <driver/adc.h>
#include <driver/i2s.h>

// This example using the DAC to generate a 10 Hz sine wave which is then sampled by the ADC
// Connect GPIO25 (DAC output) to GPIO32 (ADC1 Channel4 input)
// The raw ADC values are logged to Serial at 921600. You can redirect the output to file
// with the following command in a platformIO terminal:
// pio device monitor > log.txt
// Collect about 5s of data, edit off the first few non-data lines from the file and plot
// the data points without connecting lines (!). If the problem shows up there will be
// horizontal stripes in the data where no samples are present.

#define NSAMPLES 1024
#define SAMPLE_RATE (10000)

// The channel to read from
#define ADC_CHANNEL ADC1_CHANNEL_4  // gpio 32

// storage for the samples
uint16_t samples[NSAMPLES] = {0};

// storage for the DAC waveform at 10Hz
const uint32_t nWavePoints = SAMPLE_RATE/10;
uint16_t wave[nWavePoints];

/** Set up the I2S for both ADC and DAC
 */
void initI2S(void)
{
    i2s_config_t i2s_config = {
        .mode = (i2s_mode_t)(I2S_MODE_MASTER | I2S_MODE_RX | I2S_MODE_ADC_BUILT_IN | I2S_MODE_TX | I2S_MODE_DAC_BUILT_IN),
        .sample_rate = SAMPLE_RATE,
        .bits_per_sample = I2S_BITS_PER_SAMPLE_16BIT,
        .channel_format = I2S_CHANNEL_FMT_ONLY_RIGHT,  // stops the sample rate being doubled vs RIGHT_LEFT
        .communication_format = I2S_COMM_FORMAT_I2S_LSB,
        .intr_alloc_flags = 0,
        .dma_buf_count = 4,
        .dma_buf_len = 1024,
        .use_apll = 1,
        .tx_desc_auto_clear = 0,
        .fixed_mclk = 1
    };

    //install and start i2s driver
    i2s_driver_install(I2S_NUM_0, &i2s_config, 0, NULL);

    //init ADC pad
    i2s_set_adc_mode(ADC_UNIT_1, ADC_CHANNEL);

    // init DAC
    i2s_set_dac_mode(I2S_DAC_CHANNEL_RIGHT_EN); // right is GPIO25

    // invert the ADC so that the results match those from the normal adc read api
    adc_set_data_inv(ADC_UNIT_1, true);
}

void dacTask(void *params)
{
    size_t written;
    for(;;) {
        i2s_write(I2S_NUM_0, &wave[0], 1000, &written, portMAX_DELAY);
        i2s_write(I2S_NUM_0, &wave[500], 1000, &written, portMAX_DELAY);
    }
}

void adcTask(void *params)
{
    size_t bytes_read, total_read;
    const size_t bytesNeeded = NSAMPLES * 2;
    for(;;) {
        // read a block of raw adc values
        bytes_read = 0;
        total_read = 0;
        while(total_read < bytesNeeded) {
            i2s_read(I2S_NUM_0, (void*) &samples[total_read/2], bytesNeeded - total_read, &bytes_read, portMAX_DELAY); // this will wait until enough data is ready
            total_read += bytes_read;
        }

        // for each raw value
        for(int i=0; i<NSAMPLES; i++)
        {
            // get the channel and data
            uint32_t ch = samples[i] >> 12;    // channel is in the top 4 bits
            uint32_t value = samples[i] & 0x0FFF;

            if (ch == ADC_CHANNEL) Serial.println(value);
        }
    }
}

void setup() {
    Serial.begin(921600); // For debug

    // generate a 10 Hz sine wave
    for(int i=0; i<nWavePoints; i++) {
        wave[i] = uint8_t(sinf(float(i)*2.0f*PI/nWavePoints)*64+127) << 8;
    }

    initI2S();

    // task for writing out the waveform to the DAC
    xTaskCreatePinnedToCore(dacTask, "dacTask", 3000, NULL, 30, NULL, 1);

    // task for reading from the ADC
    xTaskCreatePinnedToCore(adcTask, "adcTask", 3000, NULL, 20, NULL, 1);
}

void loop() {
    delay(100);
}